@@ -113,11 +113,12 @@ class OptimizedPredicateSet(LiteralResolver):
         the cause of failure. This is generally much slower
         that using the equivilent `evaluate`.
 
-        Returns a tuple of (Result, Matches, Info).
+        Returns a tuple of (Result, Matches, Ctx).
         Result is a boolean, Matches a list of predices
-        and info is a dictionary containing "failed" and "literals".
-        The failed key has all the failure reasons in order.
-        The literals dict contains the resolved values for all literals.
+        and ctx is the evaluation context, containing among other
+        things the failure reasons and all of the literal resolution values.
+        The failed attribute has all the failure reasons in order.
+        The literals attribute contains the resolved values for all literals.
         """
         if self.ast is None:
             self.compile_ast()
@@ -127,11 +128,11 @@ class OptimizedPredicateSet(LiteralResolver):
         self._results = results
 
         # Analyze
-        res, info  = self.ast.analyze(self, document)
+        res, ctx = self.ast.analyze(self, document)
 
         # Reset the results array and return this instance
         self._results = None
-        return res, results, info
+        return res, results, ctx
 
     def compile_ast(self):
         """
